(*
 * Copyright (C) 2006, 2007 Jeremy Michelson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 3 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * The GNU General Public License is also available from 
 * http://www.gnu.org/copyleft/gpl.txt
*)

BeginPackage["adjacentSort`"];

adjacentSort::usage = "adjacentSort[list] sorts the list using only adjacent swaps, and returns a list of both the sorted list and a sequential list of the pairs of items that if exchanged will give the sorted list.  The sequential list of pairs is the at-that-time position of the adjacent items.  For example, adjacentSort[{4,1,3}] == {{1,3,4}, {{1,2},{2,3}}} since first the {1,2} operation yields {1,4,3} and then the {2,3} operation on that object yields {1,3,4}.  adjacentSort[list, {options}] applies the options to adjacentSort[list] and adjacentSort[list, {options}, p] uses the function p to compare adjacent items (cf. Sort[list, p]).  Allowed options are justSwaps->True/False to return only the list of pairs of items to exchange and origPos->True/False to always use the original positions of the items that are swapped instead of the then-current positions.  That is, adjacentSort[{4,1,3}, {justSwaps->True, origPos->True}] == {{1,2},{1,3}}.  The Head of the list need not be List.";

adjacentGroupSort::usage = "adjacentGroupSort[list] sorts the list using only adjacent swaps, but allowing for swaps of entire ordered groups rather than treating each entry individually.  Returned is a List of two Lists (or 3 if justSwaps->False).  The first List is a List of rules, like for adjacentSort, except that now each element refers to the position of the group, not the item.  The second List is a list of the lengths of each item, in the order that they appear in the original list l.  The options are identical to those of adjacentSort, plus showGroups->True (default False) in conjunction with justSwaps->False (the default) groups the sorted list .  For example, adjacentGroupSort[{4,1,3}] yields {{1,3,4}, {{1,2}}, {1,2}} meaning that this can ordered by swapping the 1 item '4' with 2 items {1,3}.  Similarly adjacentGroupSwap[{4,5,1,2}, showGroups->True] gives {{{1,2},{4,5}}, {{1,2}}, {2,2}}."

ClearAll[origPos, justSwaps, showGroups];

origPos::usage = "origPos is an option to adjacentSort that determines whether to use the original or at-the-time current positions in the output of the sorting"

justSwaps::usage = "justSwaps is an option to adjacentSort that determines whether to return the sorted result or just the operations required to obtain it."

showGroups::usage = "showGroups is an option to adjacentGroupSort that determines whether the sorted list is returned grouped or not.";

Protect[origPos, justSwaps, showGroups]

(* adjacentSort is a utility function.

adjacentSort sorts the list, using the ordering function p, but using
a somewhat inefficient algorithm (but thanks to
http://en.wikipedia.org/wiki/Cocktail_sort for improving my bubble sort)
in which only exchanging adjacent items are exchanged.
If returnSortedList is False, then instead of returning the equivalent of
Sort[list], a sequential list of the locations of the pairs of items
that if exchanged will give the sorted list.  This sequential list of
pairs is the at-that-time position; for example,
adjacentSort[{4,1,3},False] ==={{1,2},{2,3}} since first the "{1,2}"
operation yields {1,4,3} and then the {2,3} operation on this object
yields {1,3,4}.

As a final note on the inefficiency, compare to Sort (2.4GHz Core 2 Duo,
2GB 800MHz RAM)
In[7]:=
Timing[And@@((adjacentSort[#][[1]]===Table[i,{i,8}])&/@
        Permutations[Table[i,{i,8}]])]

Out[7]=
{14.6118 Second,True}

In[8]:=
Timing[And@@((adjacentSort[#,{origPos\[Rule]True}][[1]]===Table[i,{i,8}])&/@
        Permutations[Table[i,{i,8}]])]

Out[8]=
{19.0071 Second,True}

In[9]:=
Timing[And@@((Sort[#]===Table[i,{i,8}])&/@Permutations[Table[i,{i,8}]])]

Out[9]=
{0.207969 Second,True}

(In fact, the original speed was 11.8802 seconds but then I enabled
origPos and therefore defined Swap instead of using that code in-line
in both loops.  D'oh.)

Now that I've compiled it, it is down to about 6 seconds, so only
a factor of maybe 2.5.  D'oh.  However, the removal of the idiom
of Chapter 2.13.4 of the Mathematica book has improved the time
further.  Also, on a different (slower) computer, something like
  longJob = Table[Random[Integer,{-10000,10000}],{3200}];
  AbsoluteTiming[Length[adjacentSort[longJob,{justSwaps->True}]]]
took just less than 0.9 seconds (and returned a list of over a million
pairs of swaps!);
  AbsoluteTiming[Length[originalAdjacentSort[longJob,{justSwaps->True}]]]
took over 8 1/2 hours (and counting).

*)

Begin["`Private`"]

originalAdjacentSort[list_, ops_List:{}, p_:(OrderedQ[List[##]] &)] := 
    Module[{newlist, exchangeList, positionMap, swapped, i, top,
            bottom, returnSortedList, returnOrigPos, Swap},
      (* A utility function that affects the adjacentSort local variables *)
      (* The Unevaluated's should prevent problems should justSwaps be globally
         set to something.  But it doesn't work.
         The !TrueQ ensures a boolean result regardless
         of whether the ops are well-defined. *)
      returnSortedList = TrueQ[Evaluate[Unevaluated[justSwaps] /. 
         Flatten[{ops, Unevaluated[justSwaps]->False}]]];
      returnOrigPos = TrueQ[Evaluate[Unevaluated[origPos] /. 
         Flatten[{ops, Unevaluated[origPos]->False}]]];
      exchangeList = {};

      (* The definition of Swap is slightly different depending on
         returnOrigPos.  Note that AppendTo is used even though the Mathematica
         manual says there are more efficient ways.  One ostensibly more
         efficient method is new = Flatten[{old, {extra}}, 1] and another
         is to use Sow/Reap.  Testing shows that the latter doesn't work across
         loop iterations--rather only in a single expression--and the former
         is actually slower. *)
      If[returnOrigPos,
         positionMap = Table[i, {i, Length[list]}];
         Swap[i_, j_] := Module[{},
             newlist = ReplacePart[newlist, 
                                   Take[newlist, {i, j}],
                                  {{i}, {j}},
                                  {{2}, {1}}];
             AppendTo[exchangeList, {positionMap[[i]], positionMap[[j]]}];
             positionMap = ReplacePart[positionMap,
                                       Take[positionMap, {i, j}],
                                       {{i}, {j}},
                                       {{2}, {1}}];
             ], (* end of Swap if returnOrigPos *)
         (* else *)
         Swap[i_, j_] := Module[{},
             newlist = ReplacePart[newlist, 
                                   Take[newlist, {i, j}],
                                  {{i}, {j}},
                                  {{2}, {1}}];
             AppendTo[exchangeList, {i, j}];
             ] (* end of Swap if !returnOrigPos *)
         ]; (* end of alternate definitions of Swap *)
      (*
         This is probably a really bad algorithm. 
         We will just go through the list, swapping adjacent items if
         necessary, until we have gone through the list without swapping
         adjacent items.
      *)
      swapped = True; (* in lieu of an Until[ ...] *)
      top = Length[list];
      bottom = 1;
      newlist = list;
      While[swapped,
        swapped = False; (* we haven't! *)
        For[i = bottom,
          i < top (* don't want i == top, (i + 1 > top!) *),
          i++,
          If[!p[newlist[[i]], newlist[[i+1]]],
            swapped = True;
            Swap[i, i+1];
            ]; (* done swapping *)
          ]; (* done with the For loop going up *)
        (* If we haven't swapped yet, we won't going down *)
        If[!swapped, Break[]];
        For[i = --top, i > bottom, i--,
          If[ !p[newlist[[i-1]], newlist[[i]]],
            swapped = True;
            Swap[i-1, i];
            ]; (* done swapping *)
          ]; (* done with the For loop going down *)
        bottom++;
        ]; (* done with the While loop *)

      If[returnSortedList, 
         Return[exchangeList],
         Return[{newlist, exchangeList}]
        ];
    ]; (* end of adjacentSort *)

(* originalAdjacentSort works, but is slow.  We can speed it up by
   compiling it.  One simple way to do this requires sorting
   numbers instead of general mathematica expressions.  That is,
   given n mathematica expressions, the sort is (almost) isomorphic to sorting
   the numbers 1..n.  (E.g. Sort[{c,a,b}] is isomorphic to Sort[{3,1,2}]).
   The isomorphism breaks down if two expression are equal, but that just means
   that the sort order is ambiguous, and is otherwise not a big deal.
   It would be nice to claim that the isomorphism is implemented by the built-in
   function Ordering--and then it would really be an isomorphism--but unfortunately,
   though Ordering returns a list of integers 1..n, they are not the isomorphic to
   the expressions, but rather are the corresponding sorted positions.  E.g.
   Ordering[{c,a,b}]->{2,3,1} since what should be the first element, namely a,
   is at position 2, etc.  I think that this just means that the result,
   with justSwaps->True, of adjacentSort on the result of Ordering is just the inverse
   of the correct result, but I'm not sure.  At any rate, this naive algorithm, namely
   that the justSwaps->True result of adjacentSort[l_] should be equivalent to
   Reverse[adjacentSort[Ordering[l, {justSwaps->True}]]] seems to fail with
   origPos->True.  So we will do a (hopefully not to slow) preprocessing that differs
   from the (presumed quick postprocessing) Reverse that will map the result of Ordering
   to give the isomorphism that we want.
   
   Incidentally, the Compile function didn't like the SetDelayed.
 *)

adjacentSort::notCompiled = "Unable to find compiled version; reverting to interpreted version.";

adjacentSortCompiled = True; (* assume *)

Check[
    Install["adjacentSort"], (* else *)
    Message[adjacentSort::notCompiled];
    adjacentSortCompiled = False;
    adjacentSort = originalAdjacentSort,
    (* messages to check, so that e.g. ClearAll doesn't give an error *)
    LinkOpen::"linke", Install::"*", LinkOpen::"*"
    ];

(* now define adjacentGroupSort in terms of adjacentSort or compiledAdjacentSort *)

(* Utility function for adjacentGroupSort, it converts a list of integers into
   a one-to-one pair of lists of singlets and lengths.  For example,
   {1,2,3,5,6,4} because {{1,5,4}, {3,2,1}} corresponding to the start and
   length of each sequence of consecutive integers *) 

trim[l_List] := 
  Module[{newList, itemLength, len = 1, prev = Null, dummy},
    {{newList}, {itemLength}} =
      Drop[
        Reap[
          If[# === prev + 1,
            len++;
            prev++,
            (*else*)
            If[prev =!= Null,
              Sow[prev - len + 1, newList];
              Sow[len, itemLength];
              ];
            (*and reset *)
            prev = #;
            len = 1;
            ] & /@ Join[l, {dummy}],
          {newList, itemLength}
          ],
        1][[1]];
  Return[{newList, itemLength}];
  ];

(* another utility function, this is the inverse of trim. *)
unTrim[l : {tl_List, lens_List}] := 
  Join @@ MapThread[Range[#1, #1 + #2 - 1] &, l];

(* a final utility function, this extracts the sorted list from the final
   and initial trim'ed list. *)
unTrimSorted[stl_List, itl_List, lens_List, origList_List, sg_:False] :=
  Module[{groups, pos=1, sgf},
    MapThread[
      (groups[#1] = Take[origList, {pos, pos+#2-1}];
      pos += #2;)&,
      {itl, lens}
    ];
    sgf = If[!TrueQ[sg], (Sequence@@#)&, Identity];
    Return[ (sgf@groups[#])& /@ stl ];
  ];

adjacentGroupSort[l_List, ops_List:{}, p_:(OrderedQ[List[##]] &)] := 
  Module[{orderings, (* how to get the canonical form of l *)
          newList, (* the canonical for of l *)
          trimmedOrderings, (* newList with groups replaced by singlets *)
          trimmedLengths, (* lengths of groups, in one-to-one correspondence
                             with entries of trimmedOrderings *)
          sortResult (* the result of adjacentSort[trimmedOrderings] *)
         },
    orderings = Ordering[l, All, p];
    newList = Table[1, {Length[l]}]; (* initialized *)
    Do[newList = ReplacePart[newList, i, orderings[[i]]], {i, Length[l]}];
    {trimmedOrderings, trimmedLengths} = trim[newList];
    sortResult = adjacentSort[trimmedOrderings, ops, p];
    Return[
      If[!TrueQ[justSwaps /. ops],
         {unTrimSorted[sortResult[[1]], trimmedOrderings, trimmedLengths, l,
                       showGroups /. ops],
          Sequence@@Drop[sortResult, 1],
          trimmedLengths},
         (* else *)
         {sortResult, trimmedLengths}
      ]
    ];
  ];

End[]

EndPackage[]
