
(* ::Package:: *)

Check[ (* progress package is optional, but JLink is not. *)
      BeginPackage["grassmannOps`", 
                   {"Global`", "adjacentSort`", "JLink`", "cache`",
                    "progress`"}],
           (* needed Global` because otherwise MacOS doesn't see
              progress`ProgressWhile because it is in Global! *)
      Print["There was a problem?"];
      If[ FreeQ[$ContextPath, "adjacentSort`"],
          (* assume read in order, and so we didn't get to progress. *)
          $Failed,
          (* otherwise, just use optional Progress *)
          ProgressWhile = While;
          Print["You can ignore a problem reading the progress` package.\n",
                "While will be used instead of ProgressWhile"];
        ];
];

(* Options; can be overridden by setting the variables *)
$grassmannOpsUsesProgressWhile = True;
$grassmannMinLengthForProgressWhile = 8;
$grassmannOpsOrderedQ = OrderedQ[List[##]]&;

$grassmannOpsOrderedQ::usage = "The ordering function to use to canonicalize
 the ** product.  By default, OrderedQ is used.";

(*
  BUGS/FEATURES:
  1. x graded does not automatically imply that f[x] is graded.
  2. Commutators of f[x] with p vanish even if the commutator of x with p
     is nonvanishing.  In other words, the commutation rule for the commutator
     of f[x] and p has to be set manually.
  3. It appears that caching slows things down dramatically.  So I have
     removed it.
*)

(*

grassmannOps.m

Derives from a package that teaches Mathematica how to manipulate
Grassmann variables
by M. Headrick
(April 2003)

Modified by Jeremy Michelson (2004; Nov.-Dec. 2006)

Functionality added, and file name modified by Jeremy Michelson, Nov. 2006,
to allow for commutators and anticommutators.

Improved, we hope, Dec. 2006.

Improved [better GPower support with fractional powers and inside commutators]
and devolved [the GPower support required removal of an optimization;
Mathematica 6 support requires a workaround because the Flat attribute
no longer works properly] July, 2007.

GPower added by J. Guffin at some point and stolen from M. Headrick's 
Dec. 2004 version.

After the package is loaded, Grassmann-valued variables and functions
must be declared using the function Fermion, e.g.

June 2007: Added complex conjugation features.  In particular,
  Conjugate[x**y] == Conjugate[y]**Conjugate[x].

 Fermion[ theta1, theta2, psi[_], chi[_], G[_] ] .
*)

Fermion::usage = "Fermion[ theta1, theta2, psi[_], chi[_], G[_] ] declares that the named variables/patterns have odd grading.  Products involving such variables should always be entered using '**'."

Boson::usage = "Boson[ x, y, f[_], ] declares that the named variables/patterns have even grading.  Products involving such variables should always be entered using '**'."

(*
Fundamental (anti)commutation relations must be subsequently declared
by using the function SetCommutator (or SetAntiCommutator), e.g.
 SetCommutator[ x, p, I]
 Fermion[psi, pi]
 SetCommutator[psi, pi, -I]
*)

SetCommutator::usage = "SetCommutator[A, B, expression] declares that A**B+/-B**A=expression, with the + sign only if A and B have been previously declared to have odd grading.  This can only work for subsequent expressions that are entered with multiplication denoted using '**'.  See also SetAntiCommutator."

Commutator::usage = "Commutator[A, B] computes the (anti)commutator of A and B according to SetCommutator and SetAntiCommutator";

SetAntiCommutator::usage = "SetAntiCommutator[A, B, expression] declares that A**B+/-B**A=expression.  The difference between SetAntiCommutator and SetCommutator is that, if the grading of A and/or B has not been previously declared the unknown one(s) is set to be odd."

(*
Note that Boson[x, p] was not called in the example; however, Grading[x] is now
nonzero and even, and x (and p) must appear in ** products in order
for expressions to be evaluated correctly taking the commutation
relations into account.  Fermion must be called so that SetCommutator knows
that it should define an anticommutator instead; alternatively,
SetAntiCommutator can be used and then Fermion need not be.
Any call to Fermion or Boson on the arguments to SetCommutator or
SetAntiCommutator trumps which function was called.  For example,
 Boson[x, p]
 SetAntiCommutator[ x, p, I]
is equivalent to
 SetCommutator[x, p, I]


Expressions involving such variables should be entered using the **
(NonCommutativeMultiply) operator, e.g.

 Phi[x_]:= f[x] + psi[x]**theta1 + chi[x]**theta2 + F[x] theta1**theta2
 G[y]:= = a chi[y] - 5 psi[y]**theta1**theta2

Here f[x], F[x], and a are ordinary bosonic quantities and can therefore
be multiplied in the ordinary way (i.e. with Times). However, **
should still be used for factors with bosonic statistics but which
involve Grassmann variables, e.g.

 ( 1 + theta1**theta2 ) ** psi[x] .

The function

 Boson[var1, var2, ...]

allows one to declare variables and functions which have bosonic
statistics but potentially involve Grassmann quantities, and instructs
Mathematica always to multiply them using **.

The function

 Grading[ expr_ ]

returns a non-negative integer: 0, if expr is a pure boson, which can
be multiplied using Times; positive and odd, if expr has fermionic
statistics; or positive and even, if expr has bosonic statistics but
involves Grassmann variables. Values of the function Grading may also
be defined explicitly; for example,

 Grading[ f[x_] ] := Grading[ x ]

defines the function f to have the same grading as its argument. Any
variable whose grading is not explicitly declared is assumed to be
purely bosonic.
*)

Grading::usage = "Grading[expr_] returns a non-negative integer: 0, if expr is a pure boson which can be multiplied using Times; positive and odd if expr has fermionic statistics; or positive and even if expr has bosonic statistics but might involve Grassmann variables, or be involved in nontrivial commutation relations.  Values of the function Grading may also be defined explicitly via e.g. Grading[f[x_]] := Grading[x]";

(*
Derivatives with respect to bosonic variables can be taken in the
usual way. Derivatives with respect to Grassmann variables should be
taken using

 GD[ expr, var ] or GD[ expr, var1, var2, ... ] .

As with bosonic derivatives, values of Grassmann derivatives may be
explicitly defined:

 Fermion[ f[_,_] ];
 Boson[ g[_] ];
 GD[ f[x_,theta_], theta_ ] := g[x]
*)

GD::usage = "GD[expr, var] calculates the derivative of expr with respect to the fermionic variable var.";

GPower::usage = "GPower[expr, n] gives the expr to the nth power with respect to the ** product.";

GExp::usage = "GExp[expr] represents the exponential of expr with respect to the ** product.";

(*
The function GPower[ expr, power ] takes a power with the ** product.
(Thanks to J. Guffin for this.)

The function

 GExpand[ expr ]

works much like Expand, but also expands out ** products.

 GCollect[ expr, {var1, var2, ...} ]

collects terms with the same power of the Grassmann variables var1,
var2, ....

-- Jeremy's additions --

*)
GDot::usage = "GDot[tensor1, tensor2, ...] calculates vector and matrix products, similar to Dot, but with respect to the ** product instead of Times.";

(*
The function

 GDot[list1, list2, ...]

works much like Dot (vector and matrix multiplication) but uses ** products.

The functions
 GetGradeds[expr1, expr2, ...]
and 
 GetFermions[expr1, expr2, ...]
return a list of respectively those exprn with nonzero/odd grading.

*)

GetGradeds::usage = "GetGradeds[expr1, expr2, ...] returns a list of those expr's with nonzero grading.";

GetFermions::usage = "GetFermions[expr1, expr2, ...] returns a list of those expr's with odd grading.";

GExpand::usage = "GExpand[expr] expands the expression, including **, like Expand.  GExpand[expr, opts] is similar.";

GCollect::usage = "GCollect[expr, { var1, var2, ...}] collects terms with the same power of var1, var2, ...";

SetReal::usage = "SetReal[x, ...] declares the listed variables to be real
                  so that Conjugate[x] == x.
                  See SetConjugate";

SetConjugate::usage = "SetConjugate[z, zBar] sets z and zBar
                       to be complex conjugate.
                       SetConjugate[{z1,z2,...},{z1Bar,z2Bar,..}] sets
                       z1 and z1Bar to be complex conjugate, z2 and z2Bar
                       complex conjugate, etc, so that Conjugate[z1Bar]==z1,
                       etc.
                       See SetReal";

$grassmannOpsPostProcess::usage = "$grassmannOpsPostProcess is a function which
    is called after NonCommutativeMultiply has commuted objects to a
    canonical ordering in order to simplify the result.  The default value
    is GExpand and it is highly recommended that GExpand remain the first
    step in any user-defined function.";

$grassmannOpsPostProcess = GExpand;

Begin["`Private`"]

(* see console.m and progress.m...It seems like this should be its own
   package! :) *)
$JavaClassDir = {ToFileName[{$UserBaseDirectory, "Java"}],
                 ToFileName[{$UserBaseDirectory, "Applications", "Java"}],
                 ToFileName[{$UserBaseDirectory, "AddOns", "Java"}],
                 ToFileName[{$UserBaseDirectory, "Kernel", "Java"}],
                 ToFileName[{DirectoryName[$Input], "Java"}]
                };

(* for use with cache.m *)
ClearCachesOf[a___] := Module[{pattern, caching},
  pattern = _?(!FreeQ[#, Alternatives[a]]&);
  (* do NOT cache Commutator!  SetCommutator will overzealously clear the
     cache, thus removing definitions!  For example,
     SetCommutator[p, x, -I] followed by SetCommutator[p, Psi, D[f[x],x]**Psi]
     would remove the Commutator[p, x] == -I *)
  caching = {NonCommutativeMultiply, GetGradeds, GetFermions, GPower, GExp};
  TruncateCache[#, 0, pattern]& /@ caching;
];

(* If these are defined later, need to use Unevaluated on lhs *)
Unprotect[Conjugate];
Conjugate[NonCommutativeMultiply[a___]] := 
  NonCommutativeMultiply@@(Conjugate/@ Reverse[{a}]);
Conjugate[GPower[a_,b_]] := GPower[Conjugate[a], Conjugate[b]];
Conjugate[GExp[a_]] := GExp[Conjugate[a]];
Protect[Conjugate];

SetReal[x___] := (
  Unprotect[Conjugate];
  (Conjugate[#] = #)& /@ {x};
  Protect[Conjugate];
  );

SetConjugate[x_,y_] := SetConjugate[{x},{y}];
SetConjugate[x_List, y_List] /; Length[x] == Length[y] := (
  Unprotect[Conjugate];
  MapThread[(Conjugate[#1] = #2; Conjugate[#2] = #1;)&, {x, y}];
  Protect[Conjugate];
  );

Grading[ a_Plus ] := Max @@ (Grading /@ (List @@ a));
Grading[ a_Times ] := Plus @@ (Grading /@ (List @@ a));
Grading[ a_^b_Integer?OddQ ] := Grading[a] * Abs[b]; (* positivity *)
(* the grading must be integer even if the power is not *)
Grading[ a_^(b_ /; !IntegerQ[b] || EvenQ[b]) ] := 2 Grading[a];
Grading[ Unevaluated[GPower[a_, b_Integer?OddQ]] ] := 
    Grading[a]*Abs[b]; (* ibid *)
Grading[ Unevaluated[GPower[a_, b_ /; !IntegerQ[b] || EvenQ[b]]] ] :=
    2; (* assume fractional powers are bosonic *)
Grading[ a_NonCommutativeMultiply ] := Plus @@ (Grading /@ (List @@ a));
Grading[ Unevaluated[Derivative[__][f_][x__]] ] := Grading[f[x]];
Grading[ Unevaluated[GD[a_, _]] ] := Grading[a] + 1;
Grading[ _ ] := 0;

Fermion[ a__ ] := Module[{}, 
                    (Grading[#] = 1) & /@ {a}; 
                    ClearCachesOf[a]; (* since results may now be wrong*)
                  ];
Boson[ a__ ] := Module[{}, 
                    (Grading[#] = 2) & /@ {a};
                    ClearCachesOf[a];
                  ];


Commutator[a_Plus, b_] := (Commutator[#, b]&)/@ a;
Commutator[a_, b_Plus] := (Commutator[a, #]&)/@ b;
(* need Liebnitz rule as well, so NonCommutativeMultiply algorithm will
   work.  It is probably easier to not implement it as a Liebnitz rule,
   however.  Also, we need to be pedantic about Times: just because 2 a has
   Head Times doesn't mean that a need have zero Grading.  As a result of both
   of these, however, it is probable that the following two definitions need
   optimizing. *)
(*
Commutator[a_NonCommutativeMultiply | a_Times, b_] := a**b -
    (-1)^(Grading[a] Grading[b]) b**a;
Commutator[a_, b_NonCommutativeMultiply | b_Times] := a**b - 
    (-1)^(Grading[a] Grading[b]) b**a;
*)
(* The optimization--required by the optimization of ** itself through
   adjacentGroupSort--is to implement the Liebnitz rule. However, this
   version is the one I was trying to avoid implementing because it is
   overly recursive. *)
(*
Commutator[(a_ b_) | (a_**b_), c_] :=
  GExpand[
    a**Commutator[b,c] + (-1)^(Grading[b] Grading[c]) Commutator[a,c]**b
  ];
Commutator[a_, (b_ c_) | (b_**c_)] :=
  GExpand[
    Commutator[a,b]**c + (-1)^(Grading[a] Grading[b]) b**Commutator[a,c]
  ];
*)
(* Liebnitz rule without recursion? *)
(* First, a useful function, not defined in Mathematica < 6 *)
If[$VersionNumber < 6,
  Accumulate[l_] := Module[{sum=0},
    Return[(sum += #)& /@ l];
  ];
];
(* And now a utility function, which Accumulates up to the current element.
   i.e. {a,b,c}->{0,a,a+b} *)
PreAccumulate[l_] := Accumulate[l]-l;

Commutator[a_NonCommutativeMultiply | a_Times, b_] :=
  Module[{cumulativeGradings, signs},
    (* need (-1)^((grading of end of a's)(grading of b)) *)
    cumulativeGradings = Reverse[PreAccumulate[Reverse[Grading/@(List@@a)]]];
    signs = ((-1)^(cumulativeGradings))^Grading[b];
    Return[
      $grassmannOpsPostProcess[
        Plus @@ 
          MapIndexed[
            (Extract[signs, #2] Take[a, #2[[1]]-1]**Commutator[#,b]**
              Drop[a, #2[[1]]])&,
            List@@a]
      ]
    ];
  ];

Commutator[a_, b_NonCommutativeMultiply | b_Times] :=
  Module[{cumulativeGradings, signs},
    (* need (-1)^((grading of beginning of b's)(grading of c)) *)
    cumulativeGradings = PreAccumulate[Grading/@(List@@b)];
    signs = ((-1)^(cumulativeGradings))^Grading[a];
    Return[
      $grassmannOpsPostProcess[
        Plus @@ 
          MapIndexed[
            (Extract[signs, #2] Take[b, #2[[1]]-1]**Commutator[a,#]**
              Drop[b, #2[[1]]])&,
            List@@b]
      ]
    ];
  ];

(* end of changes for Leibnitz *)
Commutator[1/a_, b_] := (1/a)**Commutator[b,a]**(1/a);
Commutator[b_,1/a_] := (1/a)**Commutator[a,b]**(1/a);
Commutator[_,_] := 0; (* by default *)

SetCommutator[a_, b_, 0] := (Commutator[a, b] = .; ClearCachesOf[a,b]);
SetCommutator[a_, b_, com_] := Module[{ga, gb},
  If[ Grading[a] == 0, Boson[a] ];
  If[ Grading[b] == 0, Boson[b] ];
  ga = Grading[a]; (* might have changed after first call! *)
  gb = Grading[b]; (* ibid *)
  Commutator[a, b] = com;
  Commutator[b, a] = (-1)^(ga gb+1) com;
  ClearCachesOf[a,b];
];
SetAntiCommutator[a_, b_, com_] := Module[{},
  If[ Grading[a] == 0, Fermion[a] ];
  If[ Grading[b] == 0, Fermion[b] ];
  SetCommutator[a, b, com];
];

GD[ a_, b_, c__ ] := GD[ GD[a, b], c ];
GD[ GD[a_, b_], b_] := 0 /; OddQ[Grading[b]];
GD[ a_, b_, c_ ] := If[$grassmannOpsOrderedQ[{b,c}],
                       GD[ GD[a, b], c ], 
                       (-1)^Grading[b**c] GD[ GD[a,c], b ] ];
(* End Jeremy Michelson's addition.  However, the following should be
   modified according to the 2006 modifications, to minimize recursion. *)
GD[ a_, b_ ] /; Grading[b] == 0 := D[a, b];
GD[ a_, a_ ] := 1;
GD[ a_, b_ ] := 0 /; FreeQ[a, b];
GD[ a_Plus, c_ ] := GD[#, c]& /@ a;
GD[ a_Times, c_ ] := 
  Plus @@ (MapIndexed[(GD[#1, c] ** Drop[a, #2])&, (List @@ a)]);
(* TODO: The relative signs need to be checked! *)
GD[ a_NonCommutativeMultiply, c_ ] := 
  Plus @@ (MapIndexed[( (-1)^(Grading[c] Grading[Take[a, #2[[1]]-1]]) *
                       Take[a, #2[[1]]-1] ** GD[#1, c] ** Drop[a, #2[[1]]])&,
                      (List @@ a)]);
SetAttributes[ GD, Listable];

(* Need to make sure that the chain rule for ordinary derivatives
is compatible with GD[ a_,b_] := 0 /; FreeQ[a,b]! *)
Unprotect[D];
D[Unevaluated[c_. (d_:1) ** GD[a_, g_] ** (e_:1) + f_.], b__] := 
    D[c, b] d ** GD[a, g] ** e + c D[d, b] ** GD[a, g] ** e + 
      c d ** GD[D[a, b], g] ** e + c d ** GD[a, g] ** D[e, b] + D[f, b];
Protect[D];


Unprotect[NonCommutativeMultiply];
SetAttributes[NonCommutativeMultiply, Listable];
If[$VersionNumber >= 6, (* AARGH!!! Antioptimization needed or else get
                           infinite iteration required on SetCommutator[x,p,I];
                           Commutator[x,p] *)
  ClearAttributes[NonCommutativeMultiply, Flat];
  NonCommutativeMultiply[a___, b_NonCommutativeMultiply, c___] := 
     NonCommutativeMultiply[a, Sequence@@b, c];
];

(* try to keep complicated rules, not simple ones *)
Cache[GetGradeds, -$CacheDefaultSize];

GetGradeds[a___] :=
  GetGradeds[a] = Select[{a}, Grading[#] != 0 &];

Cache[GetFermions, -$CacheDefaultSize];

GetFermions[a___] := 
  GetFermions[a] = Select[{a}, OddQ[Grading[#]] &];

(* This rule needs to be commented out because it precludes defining
   other GPower[...] things
*)
GPower[x_ /; FreeQ[x, Pattern | Blank | Condition | Alternatives] 
             && Grading[x]==0,
       n_ /; Grading[n]==0] := Power[x, n];
GPower[x_, n_Integer?Positive] := Nest[x ** # &, x, n - 1];
GPower /: Power[GPower[x_, n_?(!IntegerQ[#]&)],-1] := GPower[x, -n];
GPower[Unevaluated[GPower[x_, n_]], m_] /; 
         Grading[n] == 0 && Grading[m] == 0 :=
  GPower[x, n m];
GPower[x_, 0] := 1;
GPower[1, x_] = 1;
(* Unbelievably, JLink has defined something called Default which therefore
   overrides System`Default even though JLink`Default isn't even documented.
   Moreover, this affects Linux but not Mac OS.  Or is it 6.0.0 but not
   6.0.1?  Sheesh! *)
GPower /: System`Default[GPower, 2] = 1;
(* an attempted simplification: GPower[x**x**...**x, y] == GPower[x, n y] *)
GPower[NonCommutativeMultiply[a_, b__] /; VectorQ[{b}, (#===a)&], n_] :=
  GPower[NonCommutativeMultiply[a,b], n] =
  GPower[a, (Length[{b}]+1) n];
(* finally--hopefully--define a sensible rule for 
 * GPower[something + something which involves fermions, nonpositive integer]
 * use (a + b)^x = [(1+b a^{-1})a]^x which can be further simplified 
 * without thought if a and b commute or if x is a negative integer.
 *
 * In a perfect world, we would test that commutators/anticommutators won't
 * prevent fermion^n from being zero.  Fortunately, the canonical example,
 * [p,psi]=psi, does have (p**psi)^2 = p**(psi**p)**psi 
 * = p**(p**psi-psi)**psi = 0.  However, the other canonical example,
 * {psi1,psi2} = 1 implies (psi1**psi2)^2 = psi1**psi2 DOES give an endless
 * loop, which however, is amenable to exact computation, at least when
 * the bosons and fermions commute.
 *)
GPower[a__ + b__ /; FreeQ[{a}, x_ /; OddQ[Grading[x]]] &&
                    !FreeQ[{b}, x_ /; OddQ[Grading[x]]],
       n_ /; Grading[n]==0 && (!IntegerQ[n] || n < 0)] /; 
       (IntegerQ[n] || Commutator[Plus[a], Plus[b]] === 0 ) (* &&
        * TODO: FIX THIS TEST AND UNCOMMENT IT!  Right now it fails because:
        * * b can contain bosonic parts.
        * * the exponent in the left-hand GPower is generically too big.
            What we are trying to do is make sure that all the fermions
            in the expression square to zero.  I guess just extracting
            the fermions (but not using GetFermion which doesn't go to
            deep enough level) and making sure they all anticommute is
            sufficient.
       (Grading[GPower[Plus[b], Grading[Plus[b]] * Length[{b}] + 1]] <
           Grading[Plus[b]] * Length[{b}]) 
       *) :=
  GPower[a + b, n] =
    Module[{fermionic, bosonic, normalizedFermionic, nestFunc, i},
      (* negative integer powers can be done in general, in terms of
         the inverse.
       *)
      If[Commutator[Plus[a],Plus[b]] =!= 0 && n != -1,
         (* avoid GPower[GPower[...,],] recursion. *)
         bosonic = GPower[a + b, -1]; (* just a temporary variable *)
         Return[GPower[bosonic, -n]];
        ];
      If[TrueQ[$grassmannOpsUsesProgressWhile],
         nestFunc = ProgressNestWhileList,
         (* else *)
         nestFunc[func_, expr_, test_, rest___] :=
           NestWhileList[func, expr, test];
         SetAttributes[nestFunc, {HoldAllComplete}];
      ];

      bosonic = Plus[a];
      If[Length[{b}] > 1,
         fermionic = Select[Plus[b], !FreeQ[#, x_ /; OddQ[Grading[x]]]&];
         bosonic += (Plus[b] - fermionic);
         , (* else *)
         fermionic = b;
        ];
      bosonic = GPower[bosonic, -1];
      normalizedFermionic = fermionic ** bosonic;
      (* this is a little inefficient if a and b commute *)
      i = 0; (* the first is bosonic, the second will need to be i=1 and
                i will get incremented before the calculation *)
      Return[ (* bosonic + n bosonic * fermionic + etc. *)
        (Plus@@nestFunc[(i++; (n+1-i)/i # ** normalizedFermionic) &,
                       bosonic,
                       # =!= 0 &,
                       WindowTitle -> "GPower Progress"
                       ]) ** GPower[bosonic, -n-1]
      ];
    ];
(* this rule has lower precedence than the previous one. *)
AppendTo[DownValues[GPower],
         HoldPattern[GPower[x_, n_Integer?Negative]] :>
           Nest[(1/x) ** #&, 1/x, -n - 1]
        ];
Cache[GPower, -$CacheDefaultSize];

(* This rule, needed for compatibility with GExpand and simplications, is
   defined first to avoid evaluation of other rules during the definition! *)
NonCommutativeMultiply /: Power[NonCommutativeMultiply[a_, b__], -1] := 
  NonCommutativeMultiply @@ ((1/#)& /@ Reverse[{b}]) ** (1/a);
  (* 1/(a**b) === (1/b)**(1/a) *)

(* First make sure purely commutative stuff, for which
   OrderedQ[{}]==True, or products with only one
   noncommutative factor, come out right. *)

NonCommutativeMultiply[a___] /; Length[GetGradeds[a]] <= 1 := Times[a];

(* Next do the a ** (b c) rule with minimal recursion. *)
NonCommutativeMultiply[a___] /; !FreeQ[{a}, Times, 2] := 
  (* do not wait for recursion to pull out the factors, because this ends up
   * being problematic for commuting variables which are lexically after
   * "NonCommutativeMultiply"!
   *)
 Module[{prefactor, gradeds, retVal},
   (* while we are at it, we can enforce the Flat attribute in case 
      Mathematica has not *)
   retVal = FixedPoint[ReplacePart[#, Sequence,
                                   Position[#, Times|NonCommutativeMultiply, 2]
                                  ]&,
                       {a}];
   gradeds = GetGradeds @@ retVal;
   prefactor = Times @@ Select[retVal, !MemberQ[gradeds, #]&];
   Return[(NonCommutativeMultiply@@gradeds) (prefactor)];
 ];

(* Now the general case.   The second test in the condition
   is to ensure that e.g. 3**epsilon, which is Ordered, still gets
   simplified, without incurring infinite recursion.  The first definition
   replaces the a_ ** a_ := 0 rule; it is tempting to make it a third test in
   the condition of the general case, but that creates an endless loop. *)
(* This a_ ** a_ rule is more restrictive than that in grassmann.m, because
   it has to take into account that {psi,psi} might be nonzero. 
   The second constraint fixes the infinite recursion that results because
   otherwise, for some reason, this definition takes precedence over the
   preceding Times and Flatness of NonCommutativeMultiply rules. *)
NonCommutativeMultiply[b___, a_, a_, d___] /; OddQ[Grading[a]] && 
        FreeQ[a, Times, 1] && FreeQ[a, NonCommutativeMultiply, 1] := 
   1/2 NonCommutativeMultiply[b]**Commutator[a, a]**NonCommutativeMultiply[d];
(* This rule is an esoteric simplification, which should not take precedence
   over the previous rule and which should not result in an endless loop
   with the definition of GPower. *)
NonCommutativeMultiply[b___,
                       a_^n_. | GPower[a_, n_.], a_^m_. | GPower[a_, m_.],
                       d___] /; (!IntegerQ[n] || !IntegerQ[m] || 
                                Sign[n] != Sign[m]) :=
   Which[n+m == 0, NonCommutativeMultiply[b, d],
         Grading[a]==0 && Grading[n+m]==0, a^(n+m) NonCommutativeMultiply[b,d],
         n+m > 0 || !NumericQ[n+m], NonCommutativeMultiply[b,GPower[a, n+m],d],
         IntegerQ[n+m], NonCommutativeMultiply[b, GPower[1/a, -n-m], d],
         True, NonCommutativeMultiply[b, 1/GPower[a, -n-m], d]];

(* This will probably be incompatible with GCollect, but I seem to have
   already broken GCollect anyway. *)
NonCommutativeMultiply[a___,b_Plus,c___] := (a**#**c)& /@ b;

(* Similarly, here we have to take into account the (anti)commutation relations
   during reordering.  As a byproduct, this may be too recursive.
   It might also be interesting to reimplement this with the origPos option
   to adjacentSort set to True, and then see if it is more efficient. *)
NonCommutativeMultiply[a___] /; Length[GetGradeds[a]] != Length[{a}] :=
 Module[{prefactor,argList},
   argList = GetGradeds[a];
   prefactor = (Times @@ Select[{a}, !MemberQ[argList, #]&]);
   Return[(NonCommutativeMultiply@@argList) (prefactor)];
 ];
(* This probably breaks GCollect *)
NonCommutativeMultiply[a___] /; (!OrderedQ[GetGradeds[a],$grassmannOpsOrderedQ]
                                 || Length[GetGradeds[a]] != Length[{a}] ) :=
NonCommutativeMultiply[a] =
 Module[{prefactor, expr, exchanges, argList, sortedArgs, groupLengths,
         pair, argPair, signature, com, whileFunction, pos, tmp},
 argList = GetGradeds[a];
 signature = (-1)^Length[adjacentSort[GetFermions @@ argList,
                                      {justSwaps->True},
                                      $grassmannOpsOrderedQ]];
    (* Signature does not cut it, because might have psi**psi != 0 *)
 prefactor = (Times @@ Select[{a}, !MemberQ[argList, #]&]);
 {sortedArgs, exchanges, groupLengths} =
   adjacentGroupSort[argList, {(*showGroups->True*)}, $grassmannOpsOrderedQ];
 (*sortedArgs = (NonCommutativeMultiply@@#)&/@sortedArgs;*)
 pos = 1;
 argList =
   ((tmp = Take[argList, {pos, pos+#-1}];
     pos += #;
     NonCommutativeMultiply@@tmp)&
    ) /@ groupLengths;
     (* to learn the exchanges *)
 (* the obvious thing to do is to take the first exchange pair, and recursion.
    However, this is clearly inefficient and too recursive.  So, ... *)
 expr = (NonCommutativeMultiply @@ sortedArgs) * (signature * prefactor) ;
         (* is the final expression.  Now to add in the commutators, which
            will, however, result in recursion. *)
 whileFunction = If[TrueQ[$grassmannOpsUsesProgressWhile] 
                    && Length[argList] >= $grassmannMinLengthForProgressWhile,
                    ProgressWhile, While];
 whileFunction[ Length[exchanges] > 0,
   pair = exchanges[[1]];
   exchanges = Drop[exchanges, 1]; (* for next iteration *)
   com = Commutator @@ (argPair = Take[argList, pair]);
   If[com =!= 0, (* optimization *)
      expr += 
           NonCommutativeMultiply[ Sequence @@ Take[argList, pair[[1]]-1],
           com,
           Sequence @@ Take[argList, pair[[2]]-Length[argList]] ] (prefactor);
     ];
   prefactor *= (-1)^(Times@@(Grading/@argPair));
   argList = ReplacePart[ argList, argPair, List[#]&/@pair, {{2}, {1}} ]
		(* deliberately left off the semicolon, so this appears in
		   the dialog box *)
   ]; (* done with While for each exchange *)
   Return[$grassmannOpsPostProcess[expr]];
                          (* unfortunately, the GExpand is necessary to ensure
                             that ___**(-1)**___ propagates; fortunately, it
                             speeds things up! *)
]; (* done with the meat of NonCommutativeMultiply *)

(* Clear the NonCommutativeMultiply cache.  This is only applicable for
   a "Mathematica" cache (NonCommutativeMultiply[a] = ...).
   Truncate the complicated rules (default in the Cache package),
   because the simple rules are more likely to be used in recursion. *)
Cache[NonCommutativeMultiply];

Commutator[a_, GPower[b_, x_]] /; 
  Commutator[b, Commutator[a, b]] == 0 :=
    x Commutator[a,b]**GPower[b, x-1];
Commutator[GPower[b_,x_], a_] /;
  Commutator[b, Commutator[a, b]] == 0 :=
    x Commutator[b,a]**GPower[b, x-1];


(* we cannot evaluate GExp easily, but we can at least do some
   simplifications.
   Actually, if there is a Grassmann number in there, we can use a series
   expansion, since it terminates, but I need to think about that for the
   future.
 *)
SetAttributes[GExp, Listable];
Grading[GExp[a_]] := 2; (* not quite right.  If a is a c-number, then
                           the grading is zero, but that should be accounted
                           for in the next rule.  If a is anticommuting, then
                           Grading[GExp] should be "Indeterminate" because
                           it is a sum of alternating gradings, but we might
                           be able to get away with pretending it is even *)
GExp[a_ + b__] /; Grading[b] == 0 := Exp[b] GExp[a];
(*
(* note the CBH rule always involves commutators, and not anticommutators! *)
GExp /: GExp[a_] ** GExp[b_] /; Grading[GExpand[a**b - b**a]] == 0 :=
  Exp[GExpand[1/2 (a**b - b**a)]] GExp[GExpand[a+b]]; (* in general, CBH *)
*)
(* OR the inverse rule *)
GExp /: GExp[a_ + b_] /; 
  (EvenQ[Grading[a]] || EvenQ[Grading[b]]) && GExpand[Commutator[a,b]] == 0 :=
  GExp[a] ** GExp[b];
(*GExp /: GPower[GExp[a_], n_ /; Grading[n] == 0] := GExp[n a];*)
Commutator[GExp[a_], b_ /; Head[b] =!= GExp] /; 
  (EvenQ[Grading[a]] || EvenQ[Grading[b]]) &&
   Grading[GExpand[Commutator[a,b]]] == 0 :=
  GExpand[Commutator[a, b]] ** GExp[a]; (* in general, it is an infinite series. *)
Commutator[GExp[a_], b_ /; Head[b] =!= GExp] /; 
  OddQ[Grading[a]] && OddQ[Grading[b]] &&
  Grading[GExpand[Commutator[a,b]]] == 0 :=
  (* in this situtation, we should really be using 
   * GExp[a]**b = (b + {a,b} + ...)**GExp[-a] *)
  b**GExp[a] + b**GExp[-a] + GExpand[Commutator[a,b]]**GExp[-a];
Commutator[b_ /; Head[b] =!= GExp, GExp[a_]] /; 
  (EvenQ[Grading[a]] || EvenQ[Grading[b]]) &&
  Grading[GExpand[Commutator[a,b]]] == 0 :=
  GExp[a] ** GExpand[Commutator[b, a]]; (* in general, it is an infinite series. *)
Commutator[b_ /; Head[b] =!= GExp, GExp[a_]] /;
  OddQ[Grading[a]] && OddQ[Grading[b]] && Grading[GExpand[Commutator[a,b]]] == 0 :=
  (* in this situtation, we should really be using 
   * b**GExp[a] = GExp[-a]**(b + {b,a} + ...) *)
  GExp[a] ** b + GExp[-a] ** b + GExp[-a] ** GExpand[Commutator[b, a]];
(* we need to avoid infinite recursion here *)
Commutator[GExp[a_], GExp[b_]] /;
  (EvenQ[Grading[a]] || EvenQ[Grading[b]]) &&
   Grading[GExpand[Commutator[a,b]]] == 0 :=
  Module[{signature, order},
    If[$grassmannOpsOrderedQ[GExp[a], GExp[b]],
       signature = -1; (* i.e. second term needs a change in ordering *)
       order = {GExp[a], GExp[b]},
       (* else *) signature = 1;
       order = {GExp[b], GExp[a]}
      ];
    Return[GExpand[signature (Exp[signature GExpand[Commutator[a,b]]]-1) *
                   (NonCommutativeMultiply@@order)]];
  ];
(* GExp of a grassmann number can be evaluated by power series, since
   it presumably terminates.  Just make sure it terminates by checking
   that this isn't (boson + fermion) and that fermion^n will terminate *)
GExp[a_ /; (Head[a] =!= Plus && !FreeQ[a, x_ /; OddQ[Grading[x]]]) ||
           (Head[a] === Plus && 
            And @@ (!FreeQ[#, x_ /; OddQ[Grading[x]]]& /@ (List@@a))
           )] (* TODO: This mirrors the bad code in GPower.  Fix this when
                 that is fixed.
              /; (Grading[GPower[a, Grading[a] * Length[a] + 1]] <
                  Grading[a] * Length[a]) *) := GExp[a] =
  Module[{nestFunc, n},
    If[TrueQ[$grassmannOpsUsesProgressWhile],
       nestFunc = ProgressNestWhileList,
       (* else *)
       nestFunc[func_, expr_, test_, rest___] :=
         NestWhileList[func, expr, test];
       SetAttributes[nestFunc, {HoldAllComplete}];
    ];
    n = 0; (* gets incremented right away, just in time for first nontrivial
              term *)
    Return[Plus@@(nestFunc[(n++; 1/n #**a)&, 1, (# =!= 0)&])];
  ];
Cache[GExp];
GExp[a_ /; Grading[a] == 0] := Exp[a];
  
GExpand[a_, patt___] := 
 Expand[a //. {x_NonCommutativeMultiply :> Distribute[x]}, patt];

GCollect[a_, b_List] := 
 Inner[
  NonCommutativeMultiply,
  Outer[ NonCommutativeMultiply, Sequence @@ ({1,-#}& /@ b) ] //Flatten,
  Transpose[
   Fold[ { #1/.#2->0, -GD[#1,#2] }&, a, b ],
   Reverse[Range[Length[b]]]
   ] //Flatten
  ];
        
GDot[a_List, b_List] := Inner[NonCommutativeMultiply, a, b, Plus];
GDot[a_List, b_List, c__] := GDot[GDot[a, b], c];

End[];

EndPackage[];
